//==================================================================
//
//	File name	:	listview.js
//
//	Summary		:	Listview classes
//
//	Ceated		:	2018/12/08
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (listview) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	listview.STYLE_NONE = 0;
	listview.STYLE_DATAGRID = 1;
	listview.STYLE_SUBHEAD = 2;
	listview.COL_SORT_NONE = 0;
	listview.COL_SORT_ASC = 1;
	listview.COL_SORT_DESC = 2;
	listview.COL_SORT_BOTH = 3;
	listview.COL_ITEM_LABEL = 0;
	listview.COL_ITEM_TEXTBOX = 1;
	listview.COL_ITEM_COMBOBOX = 2;
	
	// Private variables
	let oResizingCol = null;
	let iResizingColX = 0;
	
	//==========================================================
	//
	//	Class		:	Column
	//
	//	Summary		:	Class of column
	//
	//==========================================================
	listview.Column = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sLabel,			// * Label text[String]
					iWidth,			// * Column width[int]
					iAlign,			// Text align[int]
					iSortType,		// Sort type[int]
					bResizable,		// Resizable or not
					fpSort)			// Function on sort[function]
		{
			// Check parameters
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, ""),
											"listview.js", "jaswin.listview.Column.constructor");
			
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, ""),
											"listview.js", "jaswin.listview.Column.constructor");
			if(iWidth < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_WIDTH, ""),
											"listview.js", "jaswin.listview.Column.constructor");
			
			if(iAlign)
			{
				if(typeof iAlign != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ALIGN, ""),
												"listview.js", "jaswin.listview.Column.constructor");
				
				if((iAlign != jaswin.ALIGN_LEFT)
				&& (iAlign != jaswin.ALIGN_RIGHT)
				&& (iAlign != jaswin.ALIGN_CENTER))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_ALIGN, ""),
												"listview.js", "jaswin.listview.Column.constructor");
			}
			
			if(iSortType)
			{
				if(typeof iSortType != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_SORT_TYPE,
												"listview.js", "jaswin.listview.Column.constructor");
				if((iSortType != listview.COL_SORT_NONE)
				&& (iSortType != listview.COL_SORT_ASC)
				&& (iSortType != listview.COL_SORT_DESC)
				&& (iSortType != listview.COL_SORT_BOTH))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_SORT_TYPE,
												"listview.js", "jaswin.listview.Column.constructor");
			}
			
			if(bResizable)
			{
				if(typeof bResizable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_RESIZABLE,
												"listview.js", "jaswin.listview.Column.constructor");
			}
			
			if(fpSort)
			{
				if(typeof fpSort != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ON_SORT,
												"listview.js", "jaswin.listview.Column.constructor");
			}
			
			// Set properties
			this.width = iWidth;					// Column width[int]
			if(iAlign)
				this.align = iAlign;				// Text align[int]
			else
				this.align = jaswin.ALIGN_LEFT;
			if(iSortType)
				this.sortType = iSortType;			// Sort type[int]
			else
				this.sortType = listview.COL_SORT_NONE;
			this.sort = listview.COL_SORT_NONE;		// Current sort[int]
			if(bResizable)
				this.resizable = bResizable;		// Resizable or not[boolean]
			else
				this.resizable = false;
			this.onSort = fpSort;					// Function on sort[function]
			this.element = null;					// This element[DOM Element]
			this.label = null;						// Label element[DOM Element]
			this.sortIcon = null;					// Sort icon element[DOM Element]
			this.delimiter = null;					// Delimiter element[DOM Element]
			this.listView = null;					// Parent list view[ListView class]
			
			// Create element
			this.element = document.createElement("div");
			this.element.className = "listview-header-col";
			this.element.style.width = (this.width - 1) + "px";
			this.label = document.createElement("span");
			this.label.className = "listview-header-col-label";
			this.label.innerHTML = sLabel;
			this.element.appendChild(this.label);
			if(this.sortType != listview.COL_SORT_NONE)
			{
				this.element.style.cursor = "pointer";
				this.sortIcon = document.createElement("span");
				this.element.appendChild(this.sortIcon);
			}
			
			this.delimiter = document.createElement("div");
			this.delimiter.className = "listview-header-delimiter";
		}
		
		//==================================================
		//	Enable this column
		//==================================================
		enable()
		{
			if((!this.element) || (!this.listView))
				return;
			
			this.label.className = "listview-header-col-label";
			
			if(this.sort == listview.COL_SORT_ASC)
			{
				this.sortIcon.className = "listview-header-sort-asc";
				this.element.className = "listview-header-col-sorted";
				this.delimiter.className = "listview-header-delimiter-sorted";
			}
			else if(this.sort == listview.COL_SORT_DESC)
			{
				this.sortIcon.className = "listview-header-sort-desc";
				this.element.className = "listview-header-col-sorted";
				this.delimiter.className = "listview-header-delimiter-sorted";
			}
			
			if(this.sortType != listview.COL_SORT_NONE)
			{
				this.element.style.cursor = "pointer";
				if(!this.clickEvent)
				{
					this.element.addEventListener("click", this.clickEvent = (e) => {
						this.listView.clickHeader(this);
					}, false);
				}
			}
			
			if(this.resizable)
			{
				this.delimiter.style.cursor = "col-resize";
				if(!this.startResizeCol)
				{
					this.delimiter.addEventListener("mousedown", this.startResizeCol = (e) => {
						oResizingCol = this;
						iResizingColX = e.clientX;
						
						this.listView.header.addEventListener("mouseup", endResizeCol, false);
						this.listView.header.addEventListener("mouseout", outHeader, false);
						this.listView.header.addEventListener("mousemove", resizeCol, false);
					}, false);
				}
			}
			else
				this.delimiter.style.cursor = "default";
		}
		
		//==================================================
		//	Disable this column
		//==================================================
		disable()
		{
			if((!this.element) || (!this.listView))
				return;
			
			this.label.className = "listview-header-col-label-disabled";
			this.element.className = "listview-header-col";
			this.delimiter.className = "listview-header-delimiter";
			
			if(this.sortType != listview.COL_SORT_NONE)
			{
				this.sortIcon.className = null;
				this.element.style.cursor = "default";
				
				if(this.clickEvent)
				{
					this.element.removeEventListener("click", this.clickEvent, false);
					this.clickEvent = null;
				}
			}
			
			if(this.resizable)
			{
				if(this.startResizeCol)
				{
					this.delimiter.removeEventListener("mousedown", this.startResizeCol, false);
					this.startResizeCol = null;
				}
			}
			this.delimiter.style.cursor = "default";
		}
		
		//==================================================
		//	Set column label
		//==================================================
		setLabel(sLabel)		// Label text[string]
		{
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, ""),
											"listview.js", "jaswin.listview.Column.setLabel");
			this.label.innerHTML = sLabel;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	ColumnLayout
	//
	//	Summary		:	Class of column layout
	//
	//==========================================================
	listview.ColumnLayout = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sName,			// * Item name[String]
					oRect,			// * Size and position[FixedRect]
					iType,			// * Item type[int]
					oInput,			// Input class[TextBox or ComboBox]
					sDefault,		// Default value[String]
					iAlign,			// Align[int]
					iFontSize,		// Font size by point[number]
					sColor,			// Font color[String]
					bBold,			// Bold font or not[boolean]
					bFocus)			// Focus or not[boolean]
		{
			// Check parameters
			if(typeof sName != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_NAME,
											"listview.js", "jaswin.listview.ColumnLayout.constructor");
			
			if((typeof oRect != "object")
			|| (!oRect instanceof jaswin.FixedRect))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_RECT, ""),
											"listview.js", "jaswin.listview.ColumnLayout.constructor");
			
			if(typeof iType != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ITEM_TYPE,
											"listview.js", "jaswin.listview.ColumnLayout.constructor");
			if((iType != listview.COL_ITEM_LABEL)
			&& (iType != listview.COL_ITEM_TEXTBOX)
			&& (iType != listview.COL_ITEM_COMBOBOX))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_ITEM_TYPE,
											"listview.js", "jaswin.listview.ColumnLayout.constructor");
			
			if(iType == listview.COL_ITEM_TEXTBOX)
			{
				if(!jaswin.textbox)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_INCLUDE, "textbox.js"),
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
				if((typeof oInput != "object")
				|| (!oInput instanceof jaswin.textbox.TextBox))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_INPUT,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			else if(iType == listview.COL_ITEM_COMBOBOX)
			{
				if(!jaswin.list)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_INCLUDE, "list.js"),
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
				if((typeof oInput != "object")
				|| (!oInput instanceof jaswin.list.ComboBox))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_INPUT,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(sDefault)
			{
				if(typeof sDefault != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_DEFAULT,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(iAlign)
			{
				if(typeof iAlign != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ALIGN, ""),
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
				if((iAlign != jaswin.ALIGN_LEFT)
				&& (iAlign != jaswin.ALIGN_RIGHT)
				&& (iAlign != jaswin.ALIGN_CENTER))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_ALIGN, ""),
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(iFontSize)
			{
				if(typeof iFontSize != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_FONT_SIZE,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(sColor)
			{
				if(typeof sColor != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_FONT_COLOR,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(bBold)
			{
				if(typeof bBold != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_BOLD,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			if(bFocus)
			{
				if(typeof bFocus != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_FOCUS,
												"listview.js", "jaswin.listview.ColumnLayout.constructor");
			}
			
			// Set properties
			this.name = sName;
			this.rect = oRect;
			this.type = iType;
			this.input = oInput;
			this.focus = false;
			
			if(sDefault)
				this.defaultValue = sDefault;
			else
				this.defaultValue = "";
			
			if(iAlign)
				this.align = iAlign;
			else
				this.align = jaswin.ALIGN_LEFT;
			if(iFontSize)
				this.fontSize = iFontSize;
			else
				this.fontSize = 9;
			
			if(sColor)
				this.color = sColor;
			else
				this.color = "#000000";
			
			if(bBold)
				this.bold = true;
			else
				this.bold = false;
			
			if(this.input)
			{
				this.input.hide();
				if(bFocus)
					this.focus = true;
			}
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	DataBind
	//
	//	Summary		:	Class of databind to listview
	//
	//==========================================================
	listview.DataBind = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iColNumber)	// Number of columns[int]
		{
			// Check parameters
			if(typeof iColNumber != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_COL_NUMBER,
											"listview.js", "jaswin.listview.DataBind.constructor");
			if(iColNumber <= 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_COL_NUMBER,
											"listview.js", "jaswin.listview.DataBind.constructor");
			
			// Set properties
			let iCounter;
			this.columnLayouts = new Array(iColNumber);
			for(iCounter = 0; iCounter < iColNumber; iCounter++)
				this.columnLayouts[iCounter] = new Array();
			this.hiddenParameters = new Array();
		}
		
		//==================================================
		//	Add column layout
		//==================================================
		addLayout(oLayout,		// Column layout[ColumnLayout]
					iIndex)		// Column index[int]
		{
			if((typeof oLayout != "object")
			|| (!oLayout instanceof listview.ColumnLayout))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_COL_LAYOUT,
											"listview.js", "jaswin.listview.DataBind.addLayout");
			
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, ""),
											"listview.js", "jaswin.listview.DataBind.addLayout");
			if((iIndex < 0) || (iIndex >= this.columnLayouts.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, ""),
											"listview.js", "jaswin.listview.DataBind.addLayout");
			
			this.columnLayouts[iIndex].push(oLayout);
		}
		
		//==================================================
		//	Add hidden parameter
		//==================================================
		addHidden(sHiddenName)		// Name of hidden parameter[string]
		{
			if(typeof sHiddenName != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_PARAM, ""),
											"listview.js", "jaswin.listview.DataBind.addHidden");
			
			this.hiddenParameters.push(sHiddenName);
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	Row
	//
	//	Summary		:	Class of row
	//
	//==========================================================
	listview.Row = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iWidth,			// * Row width[int]
					iHeight,		// * Row height[int]
					sSubhead,		// Subhead[string]
					bDraggable,		// Draggable or not[boolean]
					fpDragstart)	// Function on start dragging[function]
		{
			// Check parameters
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, ""),
											"listview.js", "jaswin.listview.Row.constructor");
			if(iWidth < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_WIDTH, ""),
											"listview.js", "jaswin.listview.Row.constructor");
			
			if(typeof iHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, ""),
											"listview.js", "jaswin.listview.Row.constructor");
			if(iHeight < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_HEIGHT, ""),
											"listview.js", "jaswin.listview.Row.constructor");
			
			if(sSubhead)
			{
				if(typeof sSubhead != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, ""),
												"listview.js", "jaswin.listview.Row.constructor");
			}
			else if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, ""),
												"listview.js", "jaswin.listview.Row.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, ""),
													"listview.js", "jaswin.listview.Row.constructor");
				}
			}
			
			// Set properties
			this.element = null;			// This element[DOM Element]
			this.height = iHeight;			// Row height[number]
			this.checkBox = null;			// Check box element[DOM Element]
			this.subheadBox = null;			// Subhead box element[DOM Element]
			this.columns = null;			// Array of column elements[DOM Element]
			this.selected = false;			// Selected or not[boolean]
			this.checked = false;			// Checked or not[boolean]
			this.visible = true;			// Visible or not[boolean]
			this.data = {};					// Data bound to this row[Hashtable]
			this.listView = null;			// Parent list view[ListView class]
			this.controls = new Array();	// Controls on this row[Array of Control classes]
			this.subhead = false;			// Subhead or not[boolean]
			this.subheadOpen = false;		// Subhead items opened or not[boolean]
			this.draggable = false;			// Draggable or not[boolean]
			this.onDragstart = null;		// Function on dragstart[function(Argument:ListItem class, event)]
			
			if(sSubhead)
			{
				this.subhead = true;
				this.subheadOpen = true;
			}
			else
			{
				this.draggable = bDraggable;
				this.onDragstart = fpDragstart;
			}
			
			// Create element
			this.element = document.createElement("li");
			if(this.subhead)
			{
				this.element.className = "listview-subhead";
				this.element.style.width = (iWidth - 1) + "px";
				this.element.style.height = iHeight + "px";
				
				this.subheadBox = document.createElement("span");
				this.subheadBox.className = "listview-subhead-open";
				this.element.appendChild(this.subheadBox);
				let oSpan = document.createElement("span");
				oSpan.className = "listview-subhead-label";
				oSpan.innerHTML = sSubhead;
				this.element.appendChild(oSpan);
			}
			else
			{
				this.element.className = "listview-row";
				this.element.style.width = iWidth + "px";
				this.element.style.height = iHeight + "px";
				this.columns = new Array();
			}
		}
		
		//==================================================
		//	Enable this row
		//==================================================
		enable()
		{
			if((!this.element) || (!this.listView))
				return;
			
			if(this.checkBox)
			{
				if(this.checked)
					this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_CHECKED;
				this.checkBox.childNodes[0].style.cursor = "pointer";
				if(!this.mouseDownCheckEvent)
				{
					this.checkBox.addEventListener("mousedown", this.mouseDownCheckEvent = (e) => {
						this.check();
					}, false);
				}
				if(!this.hoverCheckEvent)
				{
					this.checkBox.addEventListener("mousemove", this.hoverCheckEvent = (e) => {
						this.hoverCheckBox();
					}, false);
				}
				if(!this.outCheckEvent)
				{
					this.checkBox.addEventListener("mouseout", this.outCheckEvent = (e) => {
						this.outCheckBox();
					}, false);
				}
			}
			
			if(this.subhead)
			{
				if(this.subheadOpen)
					this.subheadBox.className = "listview-subhead-open";
				else
					this.subheadBox.className = "listview-subhead-close";
				
				if(!this.clickSubheadEvent)
				{
					this.subheadBox.addEventListener("click", this.clickSubheadEvent = (e) => {
						this.listView.clickRowSubhead(this);
					}, false);
				}
			}
			else
			{
				if(!this.clickEvent)
				{
					this.element.addEventListener("click", this.clickEvent = (e) => {
						this.listView.clickRow(this);
					}, false);
				}
				if(!this.doubleClickEvent)
				{
					this.element.addEventListener("dblclick", this.doubleClickEvent = (e) => {
						this.listView.doubleClickRow();
					}, false);
				}
				if(this.draggable)
				{
					if(!this.dragStartEvent)
					{
						this.element.draggable = true;
						this.element.addEventListener("dragstart", this.dragStartEvent = (e) => {
							this.onDragstart(this, e);
						}, false);
					}
				}
				if(this.listView.contextMenu)
				{
					if(!this.contextMenuEvent)
					{
						this.element.addEventListener("contextmenu", this.contextMenuEvent = (e) => {
							this.listView.contextMenuRow(this);
						}, false);
					}
				}
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
				this.controls[iCounter].enable();
		}
		
		//==================================================
		//	Disable this row
		//==================================================
		disable()
		{
			if((!this.element) || (!this.listView))
				return;
			
			if(this.checkBox)
			{
				if(this.checked)
					this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_CHECKED_DISABLED;
				this.checkBox.childNodes[0].style.cursor = "default";
				if(this.mouseDownCheckEvent)
				{
					this.checkBox.removeEventListener("mousedown", this.mouseDownCheckEvent, false);
					this.mouseDownCheckEvent = null;
				}
				if(this.hoverCheckEvent)
				{
					this.checkBox.removeEventListener("mousemove", this.hoverCheckEvent, false);
					this.hoverCheckEvent = null;
				}
				if(this.outCheckEvent)
				{
					this.checkBox.removeEventListener("mouseout", this.outCheckEvent, false);
					this.outCheckEvent = null;
				}
			}
			if(this.subhead)
			{
				if(this.subheadOpen)
					this.subheadBox.className = "listview-subhead-open-disabled";
				else
					this.subheadBox.className = "listview-subhead-close-disabled";
				if(this.clickSubheadEvent)
				{
					this.subheadBox.removeEventListener("click", this.clickSubheadEvent, false);
					this.clickSubheadEvent = null;
				}
			}
			else
			{
				if(this.clickEvent)
				{
					this.element.removeEventListener("click", this.clickEvent, false);
					this.clickEvent = null;
				}
				if(this.doubleClickEvent)
				{
					this.element.removeEventListener("dblclick", this.doubleClickEvent, false);
					this.doubleClickEvent = null;
				}
				if(this.draggable)
				{
					this.element.draggable = false;
					if(this.dragStartEvent)
					{
						this.element.removeEventListener("dragstart", this.dragStartEvent, false);
						this.dragStartEvent = null;
					}
				}
				if(this.contextMenuEvent)
				{
					this.element.removeEventListener("contextmenu", this.contextMenuEvent, false);
					this.contextMenuEvent = null;
				}
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
				this.controls[iCounter].disable();
		}
		
		//==================================================
		//	Select this row
		//==================================================
		select(bSelect)	// Select or not[boolean]
		{
			if(bSelect)
				this.element.className = "listview-row-focus";
			else
				this.element.className = "listview-row";
			
			this.selected = bSelect;
		}
		
		//==================================================
		//	Check this row
		//==================================================
		check()
		{
			if(!this.checkBox)
				return;
			
			if(this.checked)
			{
				this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX;
				this.checked = false;
			}
			else
			{
				this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_CHECKED;
				this.checked = true;
			}
		}
		
		//==================================================
		//	Hover checkbox
		//==================================================
		hoverCheckBox()
		{
			if(!this.checked)
				this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_HOVER;
		}
		
		//==================================================
		//	Mouse out checkbox
		//==================================================
		outCheckBox()
		{
			if(!this.checked)
				this.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX;
		}
		
		//==================================================
		//	Change opened or closed
		//==================================================
		changeOpen()
		{
			if(this.subheadOpen)
			{
				this.subheadBox.className = "listview-subhead-close";
				this.subheadOpen = false;
			}
			else
			{
				this.subheadBox.className = "listview-subhead-open";
				this.subheadOpen = true;
			}
			
			return this.subheadOpen;
		}
		
		//==================================================
		//	Set row height
		//==================================================
		setHeight(iHeight)	// Row height[int]
		{
			if(typeof iHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, ""),
											"listview.js", "jaswin.panel.Row.setHeight");
			
			if((!this.columns) && (!this.subheadBox))
				return;
			
			this.height = iHeight;
			this.element.style.height = this.height + "px";
			
			if(this.checkBox)
			{
				let oSpan = this.checkBox.getElementsByTagName("span")[0];
				this.checkBox.style.height = this.height + "px";
				oSpan.style.top = Math.floor((this.height - 11) / 2) + "px";
			}
			
			if(this.columns)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
					this.columns[iCounter].style.height = this.height + "px";
			}
		}
		
		//==================================================
		//	Set label text
		//==================================================
		setLabel(iCol,		// Column index[int]
				sLabel)		// Label text[string]
		{
			if(typeof iCol != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_COL_NUMBER,
											"listview.js", "jaswin.listview.Row.setLabel");
			if((iCol < 0) || (iCol >= this.columns.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_COL_NUMBER,
											"listview.js", "jaswin.listview.Row.setLabel");
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, ""),
											"listview.js", "jaswin.listview.Row.setLabel");
			
			let oDivLabel;
			let oSpan;
			
			oDivLabel = document.createElement("div");
			oDivLabel.className = "listview-row-col-label";
			this.columns[iCol].appendChild(oDivLabel);
			oSpan = document.createElement("span");
			oSpan.style.position = "absolute";
			oSpan.style.left = "1px";
			oSpan.style.right = "1px";
			oSpan.style.top = "1px";
			oSpan.style.bottom = "1px";
			oSpan.style.lineHeight = this.height - 2 + "px";
			oSpan.innerHTML = sLabel;
			oDivLabel.appendChild(oSpan);
		}
		
		//==================================================
		//	Set control
		//==================================================
		setControl(iCol,		// Column index[int]
					oControl)	// Control[Control]
		{
			if(typeof iCol != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_COL_NUMBER,
											"listview.js", "jaswin.listview.Row.setControl");
			if((iCol < 0) || (iCol >= this.columns.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_COL_NUMBER,
											"listview.js", "jaswin.listview.Row.setControl");
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, ""),
											"listview.js", "jaswin.listview.Row.setControl");
			
			if((!this.listView) || (!this.listView.parent))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_CONTROL_ADD, ""),
											"listview.js", "jaswin.listview.Row.setControl");
			
			let oDivInput;
			
			oDivInput = document.createElement("div");
			oDivInput.className = "listview-row-col-input";
			this.columns[iCol].appendChild(oDivInput);
			oDivInput.appendChild(oControl.element);
			oControl.setParent(this.listView);
			this.controls.push(oControl);
		}
		
		//==================================================
		//	Add data to row
		//==================================================
		addData(sKey,		// Key of data[string]
				sValue)		// Value of data[string]
		{
			if(typeof sKey != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_KEY,
											"listview.js", "jaswin.listview.Row.addData");
			if(typeof sValue != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_VALUE,
											"listview.js", "jaswin.listview.Row.addData");
			
			this.data[sKey] = sValue;
		}
		
		//==================================================
		//	Add context menu event
		//==================================================
		addContextMenuEvent()
		{
			if(!this.contextMenuEvent)
			{
				this.element.addEventListener("contextmenu", this.contextMenuEvent = (e) => {
					this.listView.contextMenuRow(this);
				}, false);
			}
		}
		
		//==================================================
		//	Remove context menu event
		//==================================================
		removeContextMenuEvent()
		{
			if(this.contextMenuEvent)
			{
				this.element.removeEventListener("contextmenu", this.contextMenuEvent, false);
				this.contextMenuEvent = null;
			}
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	ListView
	//
	//	Summary		:	Class of listview
	//
	//==========================================================
	listview.ListView = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					oRect,			// * Size and position[ExpandableRect]
					iStyle,			// * Item style[int]
					oaColumns,		// * Columns[Array of Column]
					iRowHeight,		// * Default row height[int]
					bEnabled,		// * Enabled or not[boolean]
					bShowHeader,	// * Show header or not[boolean]
					bCheckbox,		// * Has checkbox or not[boolean]
					iTabIndex,		// Tab index[int]
					oDataBind,		// Databind[DataBind]
					fpChange,		// On selected index changed[function]
					fpDblClick,		// On double click function[function]
					bDraggable,		// Draggable or not[boolean]
					fpDragstart,	// Function on start dragging[function]
					bDroppable,		// Droppable or not[boolean]
					iDropType,		// Drop type[int]
					fpDrop)			// Function on dropped[function]
		{
			// Base class
			super(sID, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			if((iStyle != listview.STYLE_NONE)
			&& (iStyle != listview.STYLE_DATAGRID)
			&& (iStyle != listview.STYLE_SUBHEAD))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			if(!Array.isArray(oaColumns))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_COLUMNS, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			let iCnt = oaColumns.length;
			let iCounter;
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				if((typeof oaColumns[iCounter] != "object")
				|| (!oaColumns[iCounter] instanceof listview.Column))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_COLUMNS, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
			}
			
			if(typeof iRowHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ROW_HEIGHT, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			if(typeof bShowHeader != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SHOW_HEADER, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			if(typeof bCheckbox != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CHECKBOX, sID),
											"listview.js", "jaswin.listview.ListView.constructor");
			
			if(iStyle == listview.STYLE_DATAGRID)
			{
				if((typeof oDataBind != "object")
				|| (!oDataBind instanceof listview.DataBind))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DATABIND, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
				if(oDataBind.columnLayouts.length != oaColumns.length)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_COL_NUMBER, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
			}
			
			if(fpDblClick)
			{
				if(typeof fpDblClick != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DBLCLICK, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
			}
			
			if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, sID),
													"listview.js", "jaswin.listview.ListView.constructor");
				}
			}
			
			if(bDroppable)
			{
				if(typeof bDroppable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROPPABLE, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
				
				if(typeof iDropType != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROP_TYPE, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
				if((iDropType != jaswin.DROP_TEXT)
				&& (iDropType != jaswin.DROP_URL)
				&& (iDropType != jaswin.DROP_HTML)
				&& (iDropType != jaswin.DROP_FILE))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_DROP_TYPE, sID),
												"listview.js", "jaswin.listview.ListView.constructor");
				
				if(fpDrop)
				{
					if(typeof fpDrop != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DROP, sID),
													"listview.js", "jaswin.listview.ListView.constructor");
				}
			}
			
			// Set properties
			this.style = iStyle;					// ListView style[int]
			this.columns = oaColumns;				// Columns[Array of Column]
			this.rowWidth = 0;						// Row width[int]
			this.rowHeight = iRowHeight;			// Row height[int]
			if(this.style == listview.STYLE_SUBHEAD)
				this.checkBox = false;
			else
				this.checkBox = bCheckbox;			// Has checkbox or not[boolean]
			this.header = null;						// Header element[DOM Element]
			this.checkHeader = null;				// Checkbox header[DOM Element]
			this.rows = new Array();				// Rows[Array of Row]
			this.selectedIndex = -1;				// Index of selected row[int]
			this.sortIndex = -1;					// Index of sorted column[int]
			this.sortType = listview.COL_SORT_NONE;	// Sort type[int]
			this.dataBind = oDataBind;				// Databind[DataBind]
			this.onChange = fpChange;				// On selected index changed[function]
			this.onDoubleClick = fpDblClick;		// On double click function[function]
			this.draggable = bDraggable;			// Draggable or not[boolean]
			this.onDragstart = fpDragstart;			// Function on dragstart[function(Argument:ListItem class, event)]
			this.droppable = false;					// Droppable or not[boolean]
			this.dropType = iDropType;				// Drop type[int]
			this.onDrop = null;						// Function on drop[function(Argument:ListView class, e)]
			
			if(bDroppable)
			{
				this.droppable = bDroppable;
				this.onDrop = fpDrop;
			}
			
			// Create element
			this.element = document.createElement("div");
			this.element.className = "listview";
			if(!bShowHeader)
				this.element.style.borderRadius = "0px";
			this.focusElement = this.element;
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.width < this.minWidth)
					this.rect.width = this.minWidth;
				
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width - 2 + "px";		// 2:border
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height - 2 + "px";	// 2:border
			}
			
			// Create header
			let oDiv;
			
			this.header = document.createElement("div");
			this.header.className = "listview-header";
			
			if(this.checkBox)
			{
				// Header for checkbox
				this.checkHeader = document.createElement("div");
				this.checkHeader.className = "listview-header-col";
				this.checkHeader.style.width = "14px";
				this.checkHeader.style.cursor = "pointer";
				this.header.appendChild(this.checkHeader);
				
				oDiv = document.createElement("div");
				oDiv.className = "listview-header-delimiter";
				oDiv.style.cursor = "default";
				this.header.appendChild(oDiv);
				
				this.rowWidth += 15 + 1;
			}
			
			if(this.style == listview.STYLE_SUBHEAD)
			{
				// Header for brank
				this.checkHeader = document.createElement("div");
				this.checkHeader.className = "listview-header-col";
				this.checkHeader.style.width = "14px";
				this.header.appendChild(this.checkHeader);
				
				oDiv = document.createElement("div");
				oDiv.className = "listview-header-delimiter";
				oDiv.style.cursor = "default";
				this.header.appendChild(oDiv);
				
				this.rowWidth += 15 + 1;
			}
			
			let sBrowser = jaswin.BROWSER;
			let bResizable;
			let iLoop;
			
			iCnt = this.columns.length;
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				this.header.appendChild(this.columns[iCounter].element);
				this.header.appendChild(this.columns[iCounter].delimiter);
				this.columns[iCounter].listView = this;
				
				if(this.style == listview.STYLE_DATAGRID)
				{
					for(iLoop = 0; iLoop < this.dataBind.columnLayouts[iCounter].length; iLoop++)
					{
						if((this.dataBind.columnLayouts[iCounter][iLoop].type == listview.COL_ITEM_TEXTBOX)
						|| (this.dataBind.columnLayouts[iCounter][iLoop].type == listview.COL_ITEM_COMBOBOX))
						{
							this.columns[iCounter].resizable = false;
							this.dataBind.columnLayouts[iCounter][iLoop].input.setChangeEvent(changeListViewInputEvent);
							break;
						}
					}
				}
				
				if(this.enabled)
					this.columns[iCounter].enable();
				else
					this.columns[iCounter].disable();
				
				this.rowWidth += this.columns[iCounter].width + 1;
			}
			oDiv = document.createElement("div");
			oDiv.style.clear = "both";
			this.header.appendChild(oDiv);
			
			this.header.style.left = "0px";
			this.header.style.right = "0px";
			this.header.style.top = "0px";
			if(sBrowser == "edge")
				this.header.style.minWidth = (this.rowWidth + 1) + "px";
			else
				this.header.style.minWidth = (this.rowWidth + 18) + "px";
			this.element.appendChild(this.header);
			
			if(!bShowHeader)
				this.header.style.display = "none";
			
			// Create data area
			this.container = document.createElement("ul");
			this.container.className = "listview-data";
			this.container.style.left = "0px";
			this.container.style.right = "0px";
			if(bShowHeader)
				this.container.style.top = "17px";
			else
				this.container.style.top = "0px";
			this.container.style.bottom = "0px";
			if(sBrowser == "edge")
				this.container.style.minWidth = (this.rowWidth + 1) + "px";
			else
				this.container.style.minWidth = (this.rowWidth + 18) + "px";
			this.element.appendChild(this.container);
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.element.addEventListener("keydown", this.keyEvent = (e) => {
					return this.keyDownList(e);
				}, false);
				this.container.addEventListener("click", this.clickContainerEvent = (e) => {
					this.clickContainer(e);
				}, false);
				
				if(this.checkBox)
				{
					this.checkHeader.addEventListener("click", this.clickCheckboxHeaderEvent = (e) => {
						this.clickCheckboxHeader();
					}, false);
				}
				
				if(this.droppable)
				{
					this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
						let iCounter;
						let bFound = false;
						
						for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
						{
							if((this.dropType == jaswin.DROP_TEXT)
							&& (e.dataTransfer.types[iCounter] == "text/plain"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_URL)
							&& (e.dataTransfer.types[iCounter] == "text/uri-list"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_HTML)
							&& (e.dataTransfer.types[iCounter] == "text/html"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_FILE)
							&& (e.dataTransfer.types[iCounter] == "Files"))
								bFound = true;
						}
						if(!bFound)
							return;
						
						e.dataTransfer.dropEffect = "copy";
						e.stopPropagation();
						e.preventDefault();
					}, false);
					this.element.addEventListener("drop", this.dropEvent = (e) => {
						e.stopPropagation();
						e.preventDefault();
						if(this.onDrop)
							this.onDrop(this, e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			let iCnt;
			let iCounter;
			let iLoop;
			
			if(this.style == listview.STYLE_DATAGRID)
			{
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
				{
					iCnt = this.dataBind.columnLayouts[iCounter].length;
					for(iLoop = 0; iLoop < iCnt; iLoop++)
					{
						if(this.dataBind.columnLayouts[iCounter][iLoop].type != listview.COL_ITEM_LABEL)
							this.dataBind.columnLayouts[iCounter][iLoop].input.destroy();
					}
				}
			}
			
			this.clear();
			
			super.destroy();
		}
		
		//==================================================
		//	Set parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			
			// Input controls
			if(this.style == listview.STYLE_DATAGRID)
			{
				let iCounter;
				let iCnt;
				let iLoop;
				
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
				{
					iCnt = this.dataBind.columnLayouts[iCounter].length;
					for(iLoop = 0; iLoop < iCnt; iLoop++)
					{
						if(this.dataBind.columnLayouts[iCounter][iLoop].type != listview.COL_ITEM_LABEL)
						{
							this.element.appendChild(this.dataBind.columnLayouts[iCounter][iLoop].input.element);
							this.dataBind.columnLayouts[iCounter][iLoop].input.setParent(this);
						}
					}
				}
			}
		}
		
		//==================================================
		//	Enable this listview
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				if(!this.keyEvent)
				{
					this.element.addEventListener("keydown", this.keyEvent = (e) => {
						return this.keyDownList(e);
					}, false);
				}
				if(!this.clickContainerEvent)
				{
					this.container.addEventListener("click", this.clickContainerEvent = (e) => {
						this.clickContainer(e);
					}, false);
				}
				
				if(this.checkBox)
				{
					this.checkHeader.style.cursor = "pointer";
					if(!this.clickCheckboxHeaderEvent)
					{
						this.checkHeader.addEventListener("click", this.clickCheckboxHeaderEvent = (e) => {
							this.clickCheckboxHeader();
						}, false);
					}
				}
				
				if(this.droppable)
				{
					if(!this.dragOverEvent)
					{
						this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
							let iCounter;
							let bFound = false;
							
							for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
							{
								if((this.dropType == jaswin.DROP_TEXT)
								&& (e.dataTransfer.types[iCounter] == "text/plain"))
									bFound = true;
								else if((this.dropType == jaswin.DROP_URL)
								&& (e.dataTransfer.types[iCounter] == "text/uri-list"))
									bFound = true;
								else if((this.dropType == jaswin.DROP_HTML)
								&& (e.dataTransfer.types[iCounter] == "text/html"))
									bFound = true;
								else if((this.dropType == jaswin.DROP_FILE)
								&& (e.dataTransfer.types[iCounter] == "Files"))
									bFound = true;
							}
							if(!bFound)
								return;
							e.dataTransfer.dropEffect = "copy";
							e.stopPropagation();
							e.preventDefault();
						}, false);
					}
					if(!this.dropEvent)
					{
						this.element.addEventListener("drop", this.dropEvent = (e) => {
							e.stopPropagation();
							e.preventDefault();
							if(this.onDrop)
								this.onDrop(this, e);
						}, false);
					}
				}
				
				let iCounter;
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
					this.columns[iCounter].enable();
				
				for(iCounter = 0; iCounter < this.rows.length; iCounter++)
					this.rows[iCounter].enable();
			}
		}
		
		//==================================================
		//	Disable this listview
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			if(this.keyEvent)
			{
				this.element.removeEventListener("keydown", this.keyEvent, false);
				this.keyEvent = null;
			}
			if(this.clickContainerEvent)
			{
				this.container.removeEventListener("click", this.clickContainerEvent, false);
				this.clickContainerEvent = null;
			}
			
			if(this.checkBox)
			{
				this.checkHeader.style.cursor = "default";
				if(this.clickCheckboxHeaderEvent)
				{
					this.checkHeader.removeEventListener("click", this.clickCheckboxHeaderEvent, false);
					this.clickCheckboxHeaderEvent = null;
				}
			}
			
			if(this.dragOverEvent)
			{
				this.element.removeEventListener("dragover", this.dragOverEvent, false);
				this.dragOverEvent = null;
			}
			if(this.dropEvent)
			{
				this.element.removeEventListener("drop", this.dropEvent, false);
				this.dropEvent = null;
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
				this.columns[iCounter].disable();
			
			this.selectRow(-1);
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
				this.rows[iCounter].disable();
		}
		
		//==================================================
		//	Blur from this listview
		//==================================================
		blur()
		{
			if(this.blurEvent)
			{
				global.element.removeEventListener("click", this.blurEvent, false);
				this.blurEvent = null;
			}
			
			if(this.style == listview.STYLE_DATAGRID)
			{
				let iCounter;
				let iLoop;
				for(iCounter = 0; iCounter < this.rows.length; iCounter++)
				{
					for(iLoop = 0; iLoop < this.columns.length; iLoop++)
					{
						if(this.rows[iCounter].columns[iLoop].childNodes.length == 2)
							this.rows[iCounter].columns[iLoop].childNodes[1].style.display = "none";
					}
				}
			}
		}
		
		//==================================================
		//	Set column label
		//==================================================
		setColumnLabel(iCol, sLabel)
		{
			if(typeof iCol != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_COL_NUMBER,
											"listview.js", "jaswin.listview.ListView.setColumnLabel");
			if((iCol < 0) || (iCol >= this.columns.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_COL_NUMBER,
											"listview.js", "jaswin.listview.ListView.setColumnLabel");
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, this.id),
											"listview.js", "jaswin.listview.ListView.setColumnLabel");
			
			this.columns[iCol].setLabel(sLabel);
		}
		
		//==================================================
		//	Add empty row
		//==================================================
		addRow(sSubhead)
		{
			if(sSubhead)
			{
				if(this.style != listview.STYLE_SUBHEAD)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
												"listview.js", "jaswin.listview.ListView.addRow");
				if(typeof sSubhead != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, this.id),
												"listview.js", "jaswin.listview.ListView.addRow");
			}
			
			let oSpan;
			let oRow;
			let oDiv;
			
			if(sSubhead)
				oRow = new listview.Row(this.rowWidth, this.rowHeight, sSubhead);
			else
				oRow = new listview.Row(this.rowWidth, this.rowHeight, null, this.draggable, this.onDragstart);
			
			if(this.checkBox)
			{
				// Checkbox
				oRow.checkBox = document.createElement("div");
				oRow.checkBox.className = "listview-row-col";
				oRow.checkBox.style.width = "15px";
				oRow.checkBox.style.height = this.rowHeight + "px";
				
				oSpan = document.createElement("span");
				oSpan.className = jaswin.ICON_CHECKBOX;
				oSpan.style.left = "2px";
				oSpan.style.top = Math.floor((this.rowHeight - 11) / 2) + "px";
				if(!this.enabled)
					oSpan.style.cursor = "default";
				oRow.checkBox.appendChild(oSpan);
				oRow.element.appendChild(oRow.checkBox);
			}
			
			if(this.style == listview.STYLE_SUBHEAD)
			{
				if(!sSubhead)
				{
					// For dummy column
					oDiv = document.createElement("div");
					oDiv.className = "listview-subhead-col";
					oDiv.style.height = this.rowHeight + "px";
					oRow.element.appendChild(oDiv);
				}
			}
			
			let iCnt = this.columns.length;
			let iCounter;
			
			if(!sSubhead)
			{
				for(iCounter = 0; iCounter < iCnt; iCounter++)
				{
					// Make each column
					oRow.columns[iCounter] = document.createElement("div");
					oRow.columns[iCounter].className = "listview-row-col";
					oRow.columns[iCounter].style.width = this.columns[iCounter].width + "px";
					oRow.columns[iCounter].style.height = this.rowHeight + "px";
					if(this.columns[iCounter].align == jaswin.ALIGN_RIGHT)
						oRow.columns[iCounter].style.textAlign = "right";
					else if(this.columns[iCounter].align == jaswin.ALIGN_CENTER)
						oRow.columns[iCounter].style.textAlign = "center";
					oRow.element.appendChild(oRow.columns[iCounter]);
				}
			}
			
			oDiv = document.createElement("div");
			oDiv.style.clear = "both";
			oRow.element.appendChild(oDiv);
			
			this.rows.push(oRow);
			this.container.appendChild(oRow.element);
			oRow.listView = this;
			
			if((this.enabled) && (this.parentEnabled))
				oRow.enable();
			
			return oRow;
		}
		
		//==================================================
		//	Add data row to data area
		//==================================================
		addTextData(saData)		// * Data array[Array of string]
		{
			if((this.style != listview.STYLE_NONE)
			&& (this.style != listview.STYLE_SUBHEAD))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
											"listview.js", "jaswin.listview.ListView.addTextData");
			if(!Array.isArray(saData))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.addTextData");
			let iCnt;
			let iCounter;
			
			iCnt = saData.length;
			if(iCnt != this.columns.length)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_NUMBER_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.addTextData");
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				if(typeof saData[iCounter] != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
												"listview.js", "jaswin.listview.ListView.addTextData");
			}
			
			let oRow = this.addRow();
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				// Set text data
				oRow.setLabel(iCounter, saData[iCounter]);
			}
			
			return oRow;
		}
		
		//==================================================
		//	Add data row to data area by XML
		//==================================================
		addXmlData(oXML)
		{
			if(this.style != listview.STYLE_DATAGRID)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
											"listview.js", "jaswin.listview.ListView.addXmlData");
			if(!oXML.getElementsByTagName)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.addXmlData");
			
			let iCnt;
			let iCounter;
			let iLoop;
			let oRow = this.addRow();
			let oDivLabel;
			let oDivInput;
			let oSpan;
			let oLayout;
			let oaItems;
			let sValue;
			let sText;
			let bChecked = false;
			
			if(this.checkBox)
			{
				// Checkbox
				oaItems = oXML.getElementsByTagName("Checked")
				if((oaItems) && (oaItems.length > 0))
				{
					if(oaItems[0].textContent.toLowerCase() == "true")
						bChecked = true;
				}
				if(bChecked)
				{
					if((this.enabled) && (this.parentEnabled))
						oRow.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_CHECKED;
					else
						oRow.checkBox.childNodes[0].className = jaswin.ICON_RADIO_CHECKED_DISABLED;
					
					oRow.checked = true;
				}
			}
			
			// Set data to columns with layout
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				// Column data
				oDivLabel = document.createElement("div");
				oDivLabel.className = "listview-row-col-label";
				oRow.columns[iCounter].appendChild(oDivLabel);
				
				oDivInput = null;
				
				iCnt = this.dataBind.columnLayouts[iCounter].length;
				for(iLoop = 0; iLoop < iCnt; iLoop++)
				{
					oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
					
					oaItems = oXML.getElementsByTagName(oLayout.name)
					if((!oaItems) || (oaItems.length == 0))
						sValue = oLayout.defaultValue;
					else
						sValue = oaItems[0].textContent;
					
					oSpan = document.createElement("span");
					oSpan.style.position = "absolute";
					
					if(oLayout.align == jaswin.ALIGN_CENTER)
					{
						oSpan.style.left = oLayout.rect.posX + "px";
						oSpan.style.right = oLayout.rect.posX + "px";
					}
					else
					{
						if(oLayout.rect.anchorX == jaswin.ANCHOR_LEFT)
							oSpan.style.left = oLayout.rect.posX + "px";
						else
							oSpan.style.right = oLayout.rect.posX + "px";
						oSpan.style.width = oLayout.rect.width + "px";
					}
					if(oLayout.rect.anchorY == jaswin.ANCHOR_TOP)
						oSpan.style.top = oLayout.rect.posY + "px";
					else
						oSpan.style.bottom = oLayout.rect.posY + "px";
					
					oSpan.style.height = oLayout.rect.height + "px";
					oSpan.style.fontSize = oLayout.fontSize + "pt";
					oSpan.style.color = oLayout.color;
					if(oLayout.bold)
						oSpan.style.fontWeight = "bold";
					
					if(oLayout.type == listview.COL_ITEM_COMBOBOX)
						sText = oLayout.input.getTextByValue(sValue);
					else
						sText = sValue;
					
					oSpan.innerHTML = sText;
					oDivLabel.appendChild(oSpan);
					
					if(oLayout.type != listview.COL_ITEM_LABEL)
					{
						if(!oDivInput)
						{
							oDivInput = document.createElement("div");
							oDivInput.className = "listview-row-col-input";
							oDivInput.style.display = "none";
							oRow.columns[iCounter].appendChild(oDivInput);
						}
					}
					
					oRow.addData(oLayout.name, sValue);
				}
			}
			
			// Hidden parameters
			let sHidden;
			
			iCnt = this.dataBind.hiddenParameters.length;
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				sHidden = this.dataBind.hiddenParameters[iCounter];
				oaItems = oXML.getElementsByTagName(sHidden);
				if((!oaItems) || (oaItems.length == 0))
					sValue = "";
				else
					sValue = oaItems[0].textContent;
				
				oRow.addData(sHidden, sValue);
			}
			
			return oRow;
		}
		
		//==================================================
		//	Add data rows to data area by XML
		//==================================================
		setXmlData(oXML, sRowTag)
		{
			if((!this.enabled) || (!this.parentEnabled))
				return;
			
			if(!oXML.getElementsByTagName)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.setXmlData");
			
			if(typeof sRowTag != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ROW_TAG, this.id),
											"listview.js", "jaswin.listview.ListView.setXmlData");
			
			let oaItems = oXML.getElementsByTagName(sRowTag)
			let iCounter;
			
			if(!oaItems)
				return;
			
			for(iCounter = 0; iCounter < oaItems.length; iCounter++)
				this.addXmlData(oaItems[iCounter]);
		}
		
		//==================================================
		//	Add data row to data area by Object(JSON)
		//==================================================
		addObjectData(oData)
		{
			if(this.style != listview.STYLE_DATAGRID)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
											"listview.js", "jaswin.listview.ListView.addObjectData");
			if(!oData)
			{
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.addObjectData");
			}
			
			let iCnt;
			let iCounter;
			let iLoop;
			let oRow = this.addRow();
			let oDivLabel;
			let oDivInput;
			let oSpan;
			let oLayout;
			let oaItems;
			let sValue;
			let sText;
			
			if(this.checkBox)
			{
				// Checkbox
				if(oData["Checked"])
				{
					if((this.enabled) && (this.parentEnabled))
						oRow.checkBox.childNodes[0].className = jaswin.ICON_CHECKBOX_CHECKED;
					else
						oRow.checkBox.childNodes[0].className = jaswin.ICON_RADIO_CHECKED_DISABLED;
					
					oRow.checked = true;
				}
			}
			
			// Set data to columns with layout
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				// Column data
				oDivLabel = document.createElement("div");
				oDivLabel.className = "listview-row-col-label";
				oRow.columns[iCounter].appendChild(oDivLabel);
				
				oDivInput = null;
				
				iCnt = this.dataBind.columnLayouts[iCounter].length;
				for(iLoop = 0; iLoop < iCnt; iLoop++)
				{
					oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
					if(oData[oLayout.name])
					{
						if(typeof oData[oLayout.name] == "string")
							sValue = oData[oLayout.name];
						else if(oData[oLayout.name].toString)
							sValue = oData[oLayout.name].toString();
						else
							sValue = oLayout.defaultValue;
					}
					else
						sValue = oLayout.defaultValue;
					
					oSpan = document.createElement("span");
					oSpan.style.position = "absolute";
					
					if(oLayout.align == jaswin.ALIGN_CENTER)
					{
						oSpan.style.left = oLayout.rect.posX + "px";
						oSpan.style.right = oLayout.rect.posX + "px";
					}
					else
					{
						if(oLayout.rect.anchorX == jaswin.ANCHOR_LEFT)
							oSpan.style.left = oLayout.rect.posX + "px";
						else
							oSpan.style.right = oLayout.rect.posX + "px";
						oSpan.style.width = oLayout.rect.width + "px";
					}
					if(oLayout.rect.anchorY == jaswin.ANCHOR_TOP)
						oSpan.style.top = oLayout.rect.posY + "px";
					else
						oSpan.style.bottom = oLayout.rect.posY + "px";
					
					oSpan.style.height = oLayout.rect.height + "px";
					oSpan.style.fontSize = oLayout.fontSize + "pt";
					oSpan.style.color = oLayout.color;
					if(oLayout.bold)
						oSpan.style.fontWeight = "bold";
					
					if(oLayout.type == listview.COL_ITEM_COMBOBOX)
						sText = oLayout.input.getTextByValue(sValue);
					else
						sText = sValue;
					
					oSpan.innerHTML = sText;
					oDivLabel.appendChild(oSpan);
					
					if(oLayout.type != listview.COL_ITEM_LABEL)
					{
						if(!oDivInput)
						{
							oDivInput = document.createElement("div");
							oDivInput.className = "listview-row-col-input";
							oDivInput.style.display = "none";
							oRow.columns[iCounter].appendChild(oDivInput);
						}
					}
					
					oRow.addData(oLayout.name, sValue);
				}
			}
			
			// Hidden parameters
			let sHidden;
			
			iCnt = this.dataBind.hiddenParameters.length;
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				sHidden = this.dataBind.hiddenParameters[iCounter];
				if(oData[sHidden])
					sValue = oData[sHidden];
				else
					sValue = "";
				
				oRow.addData(sHidden, sValue);
			}
			
			return oRow;
		}
		
		//==================================================
		//	Add data rows to data area by Object(JSON)
		//==================================================
		setObjectData(oData, sRowsName)
		{
			if((!this.enabled) || (!this.parentEnabled))
				return;
			
			if(!oData)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DATA, this.id),
											"listview.js", "jaswin.listview.ListView.setXmlData");
			
			if(typeof sRowsName != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ROW_TAG, this.id),
											"listview.js", "jaswin.listview.ListView.setXmlData");
			
			let iCounter;
			let oaRows;
			
			if((oData[sRowsName]) && (Array.isArray(oData[sRowsName])))
			{
				oaRows = oData[sRowsName];
				for(iCounter = 0; iCounter < oaRows.length; iCounter++)
					this.addObjectData(oaRows[iCounter])
			}
		}
		
		//==================================================
		//	Remove data row from data area
		//==================================================
		removeData(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"listview.js", "jaswin.listview.ListView.removeData");
			
			if((iIndex < 0) || (iIndex >= this.rows.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, ""),
											"listview.js", "jaswin.listview.ListView.removeData");
			
			let iCounter;
			let iCnt;
			
			// Remove row on focus
			if(this.selectedIndex == iIndex)
			{
				this.selectedIndex = -1;
				if(this.style == listview.STYLE_DATAGRID)
				{
					// Datagrid
					let iLoop;
					let oLayout;
					
					for(iCounter = 0; iCounter < this.columns.length; iCounter++)
					{
						iCnt = this.dataBind.columnLayouts[iCounter].length;
						for(iLoop = 0; iLoop < iCnt; iLoop++)
						{
							oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
							if(oLayout.type != listview.COL_ITEM_LABEL)
							{
								oLayout.input.hide();
								if(oLayout.input.element.parentNode != this.element)
								{
									// Save input element to Listview top
									oLayout.input.element.parentNode.style.display = "none";
									this.element.appendChild(oLayout.input.element);
								}
							}
						}
					}
				}
			}
			else if(this.selectedIndex > iIndex)
				this.selectedIndex--;
			
			// Remove direct bound controls
			let oRow = this.rows[iIndex];
			iCnt = oRow.controls.length;
			for(iCounter = 0; iCounter < iCnt; iCounter++)
				oRow.controls[iCounter].destroy();
			
			this.rows.splice(iIndex, 1);
			this.container.removeChild(oRow.element);
		}
		
		//==================================================
		//	Click column header
		//==================================================
		clickHeader(oColumn)
		{
			let iCounter = 0;
			let iIndex = -1;
			
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				if(this.columns[iCounter] == oColumn)
				{
					iIndex = iCounter;
					
					if((this.columns[iCounter].sortType == listview.COL_SORT_ASC)
					&& (this.columns[iCounter].sort == listview.COL_SORT_ASC))
						return;
					
					if((this.columns[iCounter].sortType == listview.COL_SORT_DESC)
					&& (this.columns[iCounter].sort == listview.COL_SORT_DESC))
						return;
					
					if(this.columns[iCounter].sortType == listview.COL_SORT_ASC)
					{
						this.columns[iCounter].sort = listview.COL_SORT_ASC;
						this.columns[iCounter].sortIcon.className = "listview-header-sort-asc";
						this.columns[iCounter].element.className = "listview-header-col-sorted";
						this.columns[iCounter].delimiter.className = "listview-header-delimiter-sorted";
					}
					else if(this.columns[iCounter].sortType == listview.COL_SORT_DESC)
					{
						this.columns[iCounter].sort = listview.COL_SORT_DESC;
						this.columns[iCounter].sortIcon.className = "listview-header-sort-desc";
						this.columns[iCounter].element.className = "listview-header-col-sorted";
						this.columns[iCounter].delimiter.className = "listview-header-delimiter-sorted";
					}
					else
					{
						if(this.columns[iCounter].sort == listview.COL_SORT_ASC)
						{
							this.columns[iCounter].sort = listview.COL_SORT_DESC;
							this.columns[iCounter].sortIcon.className = "listview-header-sort-desc";
							this.columns[iCounter].element.className = "listview-header-col-sorted";
							this.columns[iCounter].delimiter.className = "listview-header-delimiter-sorted";
						}
						else
						{
							this.columns[iCounter].sort = listview.COL_SORT_ASC;
							this.columns[iCounter].sortIcon.className = "listview-header-sort-asc";
							this.columns[iCounter].element.className = "listview-header-col-sorted";
							this.columns[iCounter].delimiter.className = "listview-header-delimiter-sorted";
						}
					}
				}
				else if(this.columns[iCounter].sortType)
				{
					this.columns[iCounter].sort = listview.COL_SORT_NONE;
					this.columns[iCounter].sortIcon.className = null;
					this.columns[iCounter].element.className = "listview-header-col";
					this.columns[iCounter].delimiter.className = "listview-header-delimiter";
				}
			}
			
			if(iIndex != -1)
			{
				this.sortIndex = iIndex;
				this.sortType = this.columns[iIndex].sort;
				
				if(this.columns[iIndex].onSort)
				{
					this.columns[iIndex].onSort(this.columns[iIndex]);
					this.redraw();
				}
			}
		}
		
		//==================================================
		//	Click container space
		//==================================================
		clickContainer(e)
		{
			let oElement = e.target;
			
			if(oElement != this.container)
				return;
			if(this.rows.length == 0)
				return;
			
			this.selectRow(-1);
		}
		
		//==================================================
		//	Click data row
		//==================================================
		clickRow(oRow)
		{
			let iCounter = 0;
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(this.rows[iCounter] == oRow)
				{
					if(this.rows[iCounter].subhead)
						this.selectRow(-1);
					else
						this.selectRow(iCounter);
					break;
				}
			}
		}
		
		//==================================================
		//	Select data row
		//==================================================
		selectRow(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"listview.js", "jaswin.listview.ListView.selectRow");
			if((iIndex < -1) || (iIndex >= this.rows.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, ""),
											"listview.js", "jaswin.listview.ListView.selectRow");
			
			if((iIndex != -1)
			&& (this.rows[iIndex].subhead))
				return;
			
			let iCounter;
			let iLoop;
			
			if(iIndex == this.selectedIndex)
			{
				// Selected the same row
				if(this.style == listview.STYLE_DATAGRID)
				{
					let iLoop;
					
					for(iCounter = 0; iCounter < this.rows.length; iCounter++)
					{
						for(iLoop = 0; iLoop < this.columns.length; iLoop++)
						{
							if(this.rows[iCounter].columns[iLoop].childNodes.length == 2)
								this.rows[iCounter].columns[iLoop].childNodes[1].style.display = "block";
						}
					}
				}
				
				if(!this.blurEvent)
				{
					global.element.addEventListener("click", this.blurEvent = (e) => {
						let oElement = e.target;
						if(!this.element.contains(oElement))
							this.blur();
					}, false);
				}
				return;
			}
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				// Row selection
				if(iCounter == iIndex)
					this.rows[iCounter].select(true);
				else
				{
					if(!this.rows[iCounter].subhead)
						this.rows[iCounter].select(false);
				}
			}
			this.selectedIndex = iIndex;
			
			if(this.selectedIndex != -1)
			{
				let iTop = 0;
				let iBottom = 0;
				
				for(iCounter = 0; iCounter < iIndex; iCounter++)
				{
					if(this.rows[iCounter].visible)
						iTop += this.rows[iCounter].height + 1;
				}
				iBottom = iTop + this.rows[iIndex].height + 1;
				if(this.container.scrollTop > iTop)
					this.container.scrollTop = iTop;
				else if(this.container.clientHeight < iBottom - this.container.scrollTop)
					this.container.scrollTop = iBottom - this.container.clientHeight;
			}
			
			if(this.style == listview.STYLE_DATAGRID)
			{
				// Input control
				let iCnt;
				let oLayout;
				let sValue;
				let oFocusItem = null;
				
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
				{
					iCnt = this.dataBind.columnLayouts[iCounter].length;
					for(iLoop = 0; iLoop < iCnt; iLoop++)
					{
						oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
						if(oLayout.type != listview.COL_ITEM_LABEL)
						{
							if(this.selectedIndex == -1)
							{
								oLayout.input.hide();
								if(oLayout.input.element.parentNode != this.element)
								{
									// Save input element to Listview top
									oLayout.input.element.parentNode.style.display = "none";
									this.element.appendChild(oLayout.input.element);
								}
							}
							else
							{
								this.rows[this.selectedIndex].columns[iCounter].childNodes[1].appendChild(oLayout.input.element);
								
								sValue = this.rows[this.selectedIndex].data[oLayout.name];
								if(oLayout.type == listview.COL_ITEM_COMBOBOX)
									oLayout.input.selectListByValue(sValue);
								else if(oLayout.type == listview.COL_ITEM_TEXTBOX)
									oLayout.input.setValue(sValue);
								
								oLayout.input.show();
								this.rows[this.selectedIndex].columns[iCounter].childNodes[1].style.display = "block";
								
								if((!oFocusItem) && (oLayout.focus))
									oFocusItem = oLayout.input.element;
							}
						}
					}
					
					for(iLoop = 0; iLoop < this.rows.length; iLoop++)
					{
						if((iLoop != this.selectedIndex)
						&& (this.rows[iLoop].columns[iCounter].childNodes.length == 2))
							this.rows[iLoop].columns[iCounter].childNodes[1].style.display = "none";
					}
				}
				if(oFocusItem)
					oFocusItem.focus();
			}
			
			// Event function
			if(this.selectedIndex == -1)
			{
				if(this.onChange)
					this.onChange(null);
			}
			else
			{
				if(!this.blurEvent)
				{
					global.element.addEventListener("click", this.blurEvent = (e) => {
						let oElement = e.target;
						if(!this.element.contains(oElement))
							this.blur();
					}, false);
				}
				
				if(this.onChange)
					this.onChange(this.rows[this.selectedIndex]);
			}
		}
		
		//==================================================
		//	Double click row
		//==================================================
		doubleClickRow()
		{
			if((!this.enabled) || (!this.parentEnabled))
				return;
			if(this.selectedIndex == -1)
				return;
			
			if(this.onDoubleClick)
				this.onDoubleClick(this.rows[this.selectedIndex]);
		}
		
		//==================================================
		//	Set context menu to this listview
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			let oParentWnd = jaswin.findParentWindow(this);
			oParentWnd.container.appendChild(oMenu.element);
			
			oMenu.setParent(this, 0);
			
			if((this.enabled) && (this.parentEnabled))
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.rows.length; iCounter++)
					this.rows[iCounter].addContextMenuEvent();
			}
		}
		
		//==================================================
		//	Remove context menu from this listview
		//==================================================
		removeContextMenu()
		{
			super.removeContextMenu();
			
			let iCounter;
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
				this.rows[iCounter].removeContextMenuEvent();
		}
		
		//==================================================
		//	Context menu to data row
		//==================================================
		contextMenuRow(oRow)
		{
			let iCounter = 0;
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(this.rows[iCounter] == oRow)
				{
					if(iCounter != this.selectedIndex)
						this.selectRow(iCounter);
					
					if(this.contextMenu)
					{
						if(this.onContextMenu)
						{
							if(!this.onContextMenu(this.rows[iCounter]))
							{
								jaswin.menu.menuCancel();
								break;
							}
						}
						this.focusElement.focus();
						this.contextMenu.showContext(e);
					}
					break;
				}
			}
		}
		
		//==================================================
		//	Click checkbox of data row
		//==================================================
		clickCheckboxHeader()
		{
			let iCounter;
			let bCheck = true;
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(this.rows[iCounter].checked)
				{
					bCheck = false;
					break;
				}
			}
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(this.rows[iCounter].checked != bCheck)
					this.rows[iCounter].check();
			}
		}
		
		//==================================================
		//	Click subhead box of data row
		//==================================================
		clickRowSubhead(oRow)
		{
			let iCounter = 0;
			let bOpen = false;
			let bStart = false;
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(bStart)
				{
					if(this.rows[iCounter].subhead)
						break;
					if(bOpen)
					{
						this.rows[iCounter].element.style.display = "block";
						this.rows[iCounter].visible = true;
					}
					else
					{
						this.rows[iCounter].element.style.display = "none";
						this.rows[iCounter].visible = false;
						if(iCounter == this.selectedIndex)
							this.selectRow(-1);
					}
				}
				else
				{
					if(this.rows[iCounter] == oRow)
					{
						bOpen = this.rows[iCounter].changeOpen();
						bStart = true;
					}
				}
			}
		}
		
		//==================================================
		//	Clear rows
		//==================================================
		clear()
		{
			let iCounter;
			let iCnt = this.rows.length;
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
				this.removeData(0);
			
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				if(this.columns[iCounter].sortType)
				{
					this.columns[iCounter].sort = listview.COL_SORT_NONE;
					this.columns[iCounter].sortIcon.className = null;
					this.columns[iCounter].element.className = "listview-header-col";
					this.columns[iCounter].delimiter.className = "listview-header-delimiter";
				}
			}
			this.container.scrollTop = 0;
		}
		
		//==================================================
		//	Clear rows
		//==================================================
		redraw()
		{
			let iCounter;
			let oElem;
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				oElem = this.container.childNodes[iCounter];
				if(oElem != this.rows[iCounter].element)
					this.container.insertBefore(this.rows[iCounter].element, oElem);
			}
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				if(this.rows[iCounter].selected)
				{
					this.selectedIndex = iCounter;
					return;
				}
			}
			
			this.selectedIndex = -1;
		}
		
		//==================================================
		//	Change input
		//==================================================
		changeInput(oInput)
		{
			if(this.style != listview.STYLE_DATAGRID)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
											"listview.js", "jaswin.listview.ListView.changeInput");
			let iCounter;
			let iCnt;
			let iLoop;
			let oLayout;
			let sValue;
			let sText;
			
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				iCnt = this.dataBind.columnLayouts[iCounter].length;
				for(iLoop = 0; iLoop < iCnt; iLoop++)
				{
					oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
					if(oLayout.type == listview.COL_ITEM_LABEL)
						continue;
					
					if(oLayout.type == listview.COL_ITEM_TEXTBOX)
					{
						if(oLayout.input == oInput)
						{
							sValue = oInput.getValue();
							sText = oInput.getText();
							this.rows[this.selectedIndex].data[oLayout.name] = sValue;
							this.rows[this.selectedIndex].columns[iCounter].childNodes[0].childNodes[iLoop].innerHTML = sText;
							return;
						}
					}
					else if(oLayout.type == listview.COL_ITEM_COMBOBOX)
					{
						if(oLayout.input.contains(oInput))
						{
							sValue = oInput.value;
							sText = oInput.text;
							if(!sText)
								sText = "";
							
							this.rows[this.selectedIndex].data[oLayout.name] = sValue;
							this.rows[this.selectedIndex].columns[iCounter].childNodes[0].childNodes[iLoop].innerHTML = sText;
							return;
						}
					}
				}
			}
		}
		
		//==================================================
		//	Blur from row
		//==================================================
		blurRow()
		{
			if(this.selectedIndex == -1)
				return;
			
			if(this.style == listview.STYLE_DATAGRID)
			{
				let iCounter;
				let iCnt;
				let iLoop;
				let oLayout;
				let sValue;
				let sText;
				
				for(iCounter = 0; iCounter < this.columns.length; iCounter++)
				{
					iCnt = this.dataBind.columnLayouts[iCounter].length;
					for(iLoop = 0; iLoop < iCnt; iLoop++)
					{
						oLayout = this.dataBind.columnLayouts[iCounter][iLoop];
						if(oLayout.type == listview.COL_ITEM_LABEL)
							continue;
						
						if(oLayout.type == listview.COL_ITEM_TEXTBOX)
						{
							sValue = oLayout.input.getValue();
							sText = oLayout.input.getText();
							this.rows[this.selectedIndex].data[oLayout.name] = sValue;
							this.rows[this.selectedIndex].columns[iCounter].childNodes[0].childNodes[iLoop].innerHTML = sText;
							
							if(jaswin.BROWSER == "chrome")
								oLayout.input.element.blur();
						}
					}
				}
			}
		}
		
		//==================================================
		//	Resize column
		//==================================================
		resizeColumn(iGap)
		{
			let iCounter;
			let iIndex = -1;
			let iCurWidth;
			let sBrowser = jaswin.BROWSER;
			
			for(iCounter = 0; iCounter < this.columns.length; iCounter++)
			{
				if(this.columns[iCounter] == oResizingCol)
				{
					iIndex = iCounter;
					break;
				}
			}
			
			if(iIndex == -1)
				return;
			
			iCurWidth = parseInt(this.columns[iIndex].element.style.width);
			if(iCurWidth + iGap < 0)
				return;
			
			this.columns[iIndex].width = iCurWidth + iGap + 1;
			this.columns[iIndex].element.style.width = iCurWidth + iGap + "px";
			this.rowWidth += iGap;
			
			for(iCounter = 0; iCounter < this.rows.length; iCounter++)
			{
				this.rows[iCounter].element.style.width = parseInt(this.rows[iCounter].element.style.width) + iGap + "px";
				if(!this.rows[iCounter].subhead)
					this.rows[iCounter].columns[iIndex].style.width = iCurWidth + iGap + 1 + "px";
			}
			if(sBrowser == "edge")
			{
				this.header.style.minWidth = (this.rowWidth + 1) + "px";
				this.container.style.minWidth = (this.rowWidth + 1) + "px";
			}
			else
			{
				this.header.style.minWidth = (this.rowWidth + 18) + "px";
				this.container.style.minWidth = (this.rowWidth + 18) + "px";
			}
		}
		
		//==================================================
		//	Key down on this listview
		//==================================================
		keyDownList(e)
		{
			if((this.contextMenu) && (this.contextMenu.isVisible()))
				return this.contextMenu.menuKey(e);
			
			let iCounter;
			
			if(e.keyCode == 13)
			{
				// Enter
				if(this.selectedIndex == - 1)
					return true;
				
				this.doubleClickRow();
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 40)
			{
				// down
				if(this.selectedIndex != this.rows.length - 1)
				{
					if(this.selectedIndex != -1)
						this.blurRow();
					
					for(iCounter = this.selectedIndex + 1; iCounter < this.rows.length; iCounter++)
					{
						if((!this.rows[iCounter].subhead)
						&& (this.rows[iCounter].visible))
						{
							this.selectRow(iCounter);
							break;
						}
					}
				}
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 38)
			{
				// up
				if(this.selectedIndex > 0)
				{
					if(this.selectedIndex != -1)
						this.blurRow();
					
					for(iCounter = this.selectedIndex - 1; iCounter >= 0; iCounter--)
					{
						if((!this.rows[iCounter].subhead)
						&& (this.rows[iCounter].visible))
						{
							this.selectRow(iCounter);
							break;
						}
					}
				}
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 36)
			{
				// Home
				if(this.rows.length > 0)
				{
					if(this.selectedIndex != -1)
						this.blurRow();
					
					for(iCounter = 0; iCounter < this.rows.length; iCounter++)
					{
						if((!this.rows[iCounter].subhead)
						&& (this.rows[iCounter].visible))
						{
							this.selectRow(iCounter);
							break;
						}
					}
				}
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 35)
			{
				// End
				if(this.rows.length > 0)
				{
					if(this.selectedIndex != -1)
						this.blurRow();
					
					for(iCounter = this.rows.length - 1; iCounter >= 0; iCounter--)
					{
						if((!this.rows[iCounter].subhead)
						&& (this.rows[iCounter].visible))
						{
							this.selectRow(iCounter);
							break;
						}
					}
				}
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 32)
			{
				// space
				if(!this.checkBox)
					return true;
				if(this.selectedIndex == -1)
					return true;
				
				this.rows[this.selectedIndex].check();
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			return true;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	resizeCol
	//
	//	Summary		:	Event handler of resizing a column
	//
	//==========================================================
	function resizeCol(e)
	{
		let oElement = e.target;
		let iCurX = e.clientX;
		let oSelection = window.getSelection();
		let oList = null;
		
		oSelection.removeAllRanges();
		if(!oResizingCol)
			return;
		oList = oResizingCol.listView;
		
		if(!oList.header.contains(oElement))
		{
			oList.header.removeEventListener("mouseup", endResizeCol, false);
			oList.header.removeEventListener("mouseout", outHeader, false);
			oList.header.removeEventListener("mousemove", resizeCol, false);
			oResizingCol = null;
			return;
		}
		
		oList.resizeColumn(iCurX - iResizingColX);
		iResizingColX = iCurX;
	}
	
	
	//==========================================================
	//
	//	Function	:	outHeader
	//
	//	Summary		:	Event handler of mouse out header
	//
	//==========================================================
	function outHeader(e)
	{
		let oElement = e.target;
		let oList = null;
		
		if(!oResizingCol)
			return;
		oList = oResizingCol.listView;
		
		if(oElement == oList.header)
		{
			oList.header.removeEventListener("mouseup", endResizeCol, false);
			oList.header.removeEventListener("mouseout", outHeader, false);
			oList.header.removeEventListener("mousemove", resizeCol, false);
			oResizingCol = null;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	endResizeCol
	//
	//	Summary		:	Event handler of end resizing a column
	//
	//==========================================================
	function endResizeCol(e)
	{
		let oList = null;
		
		if(!oResizingCol)
			return;
		oList = oResizingCol.listView;
		
		oList.header.removeEventListener("mouseup", endResizeCol, false);
		oList.header.removeEventListener("mouseout", outHeader, false);
		oList.header.removeEventListener("mousemove", resizeCol, false);
		oResizingCol = null;
	}
	
	
	//==========================================================
	//
	//	Function	:	changeListViewInputEvent
	//
	//	Summary		:	Event handler of change a input
	//
	//==========================================================
	function changeListViewInputEvent(oInput)
	{
		let oTarget = oInput;
		
		if(oInput instanceof jaswin.Control)
			oTarget = oInput;
		else
			oTarget = oInput.parent;
		
		oTarget.parent.changeInput(oInput);
	}
	
}(jaswin.listview = jaswin.listview || {}));
